import os
import json
from threading import RLock
from datetime import datetime

from .util import Utility
from .util import CustomJSONEncoder
from .error import CollectionNotFoundError
from .error import DocumentValidationError
from .error import ReservedKeyError
from .error import UnsupportedOperatorError


class Database:
    
    
    def __init__(self, db_file: str ="database") -> None:
        """Initialize the Database"""
        
        self.EXT = ".json"
        self.VERSION = "2.0.0"
        
        self.DB_FILE = db_file
        if not db_file.endswith(self.EXT):
            self.DB_FILE = db_file + self.EXT
        self.PATH = os.getcwd()
        self.LOCK = RLock()

        self.SKELETON = {"_meta": {"_version": self.VERSION, "_path": self.PATH, "_count": {}, "_schema": {}}}
        self.RESERVED_KEYS = ["_meta", "_version", "_path", "_count", "_schema", "_id_"]

        if not os.path.exists(self.DB_FILE):
            with open(self.DB_FILE, "w") as f:
                json.dump(self.SKELETON, f, indent=4)


    def _read_db(self) -> dict:
        """Read the database from the file."""
        
        with self.LOCK:
            with open(self.DB_FILE, "r") as f:
                return json.load(f)


    def _write_db(self, data: dict) -> None:
        """Write the database to the file."""
        
        with self.LOCK:
            with open(self.DB_FILE, "w") as f:
                json.dump(data, f, indent=4, cls=CustomJSONEncoder)


    def drop_db(self) -> bool:
        """Delete the entire database file."""
        
        with self.LOCK:
            if os.path.exists(self.DB_FILE):
                os.remove(self.DB_FILE)
                return True
        return False


    def list(self) -> dict:
        """Return a dict of all collections."""
        
        db = self._read_db()
        collections = list(db.keys())
        count = len(collections) - 1
        return {"collections": collections[1:], "count": count}


    def set_schema(self, collection: str, schema: dict={}) -> None:
        """Define a schema for a collection."""
        with self.LOCK:
            self._validate_collection_exists(collection)
            
            schema_str = Utility._type_to_string(schema)
                
            db = self._read_db()
            db["_meta"]["_schema"][collection] = schema_str
            self._write_db(db)


    def get_schema(self, collection: str) -> dict:
        """Retrieve the schema for a collection."""
        
        db = self._read_db()
        schema = db.get("_meta", {}).get("_schema", {}).get(collection, None)
        
        if schema is None:
            return {}
            
        return Utility._string_to_type(schema)
    
    
    def _set_count(self, collection: str) -> None:
        """Initialize or Update the count for a collection."""
        with self.LOCK:
            db = self._read_db()
            db["_meta"]["_count"][collection] = len(db[collection])
            self._write_db(db)
    
    
    def get_count(self, collection: str) -> int:
        """Retrieve the count for a collection."""
        
        self._validate_collection_exists(collection)
        
        with self.LOCK:
            db = self._read_db()
            count = db.get("_meta", {}).get("_count", {}).get(collection, 0)
            return count


    def _validate_collection_exists(self, collection: str) -> None:
        """Check if a collection exists in the database."""
        
        db = self._read_db()
        if collection in self.RESERVED_KEYS:
            raise ReservedKeyError()
        if collection not in db:
            raise CollectionNotFoundError(collection)


    def collection(self, collection: str, schema: dict = {}) -> None:
        """Create a new collection with a schema."""
        
        if collection in self.RESERVED_KEYS:
            raise ReservedKeyError()
        else:
            flag = False
            with self.LOCK:
                db = self._read_db()
                if collection not in db:
                    flag = True
                    db[collection] = []
                    self._write_db(db)
            if flag == True:                            
                self.set_schema(collection, schema)
                self._set_count(collection)


    def drop_collection(self, collection: str) -> bool:
        """Delete a collection."""

        with self.LOCK:
            try:
                self._validate_collection_exists(collection)

                db = self._read_db()
                db["_meta"]["_schema"].pop(collection, None)
                db["_meta"]["_count"].pop(collection, None)
                db.pop(collection, None)
                self._write_db(db)
                return True
                
            except Exception as e:
                raise e
        return False
        
    
    def get_collection_data(self, collection: str) -> dict:
        """Get a collection's data."""
        
        with self.LOCK:
            self._validate_collection_exists(collection)
            
            db = self._read_db()
            collection_schema = db["_meta"]["_schema"][collection]
            collection_count = db["_meta"]["_count"][collection]
            data = db[collection][:5]
            return {collection: {"_schema": collection_schema, "count": collection_count, "data": data}}


    def _validate_document(self, collection: str, document: dict) -> bool:
        """Validate a document against the collection's schema."""
        
        for key in document.keys():
            if key in self.RESERVED_KEYS:
                raise ReservedKeyError(f"'{key}' {self.RESERVED_KEYS} are reserved keys.")
        
        schema = self.get_schema(collection)
        if not schema:
            return True

        for field, field_type in schema.items():
            if field not in document:
                raise DocumentValidationError(f"Missing required field: {field}")
            
            if field_type is datetime:
                if not isinstance(document[field], (str, datetime)):
                    raise DocumentValidationError(f"Field '{field}' must be a datetime object or a string in ISO format.")
                if isinstance(document[field], str):
                    try:
                        document[field] = datetime.fromisoformat(document[field])
                    except ValueError:
                        raise DocumentValidationError(f"Field '{field}' must be a valid ISO 8601 datetime string.")
            elif not isinstance(document[field], field_type):
                raise DocumentValidationError(f"Field '{field}' must be of type {field_type.__name__}.")
        return True


    def add(self, collection: str, document: dict) -> str:
        """Add a new document to a collection."""
        
        with self.LOCK:
            self._validate_collection_exists(collection)
            
            db = self._read_db()

            try:
                self._validate_document(collection, document)
            except DocumentValidationError as e:
                raise e

            unique_id = Utility.generate_id(collection)
            document.setdefault("_id", unique_id)
            db[collection].append(document)
        
            self._write_db(db)
            
            self._set_count(collection)

            return unique_id


    def add_many(self, collection: str, documents: list) -> list:
        """Add multiple new documents to a collection."""
        
        with self.LOCK:
            self._validate_collection_exists(collection)
            
            db = self._read_db()
            added_ids = []

            for document in documents:
                try:
                    self._validate_document(collection, document)
                    unique_id = Utility.generate_id(collection)
                    document.setdefault("_id", unique_id)
                    db[collection].append(document)
                    added_ids.append(unique_id)
                except DocumentValidationError as e:
                    raise e
            
            self._write_db(db)
            self._set_count(collection)

            return added_ids


    def _evaluate_condition(self, doc_value: any, condition: dict) -> bool:
        """Evaluate a condition (support for $gt, $lt, $ne, $eq)."""
        
        if doc_value is None:
            return False
        
        if isinstance(condition, dict):
            for operator, value in condition.items():
                if operator == '$gt':
                    if not doc_value > value:
                        return False
                elif operator == '$lt':
                    if not doc_value < value:
                        return False
                elif operator == '$eq':
                    if not doc_value == value:
                        return False
                elif operator == '$ne':
                    if not doc_value != value:
                        return False
                else:
                    raise UnsupportedOperatorError(operator)
        else:
            return doc_value == condition
        return True


    def find(self, collection: str, query: dict =None, limit: int =None, skip: int =0, sort: str =None, order: str ="asc") -> list:
    
        with self.LOCK:
            self._validate_collection_exists(collection)
            db = self._read_db()

            if not query:
                documents = db[collection]
            else:
                documents = []
                for doc in db[collection]:
                    match = True
                    for k, v in query.items():
                        if k == '$or' and isinstance(v, list):
                            match = any(
                                all(self._evaluate_condition(doc.get(field), cond) for field, cond in subquery.items())
                                for subquery in v
                            )
                        elif k == '$and' and isinstance(v, list):
                            match = all(
                                all(self._evaluate_condition(doc.get(field), cond) for field, cond in subquery.items())
                                for subquery in v
                            )
                        else:
                            match = self._evaluate_condition(doc.get(k), v)

                        if not match:
                            break

                    if match:
                        documents.append(doc)

            if sort:
                reverse = order.lower() == "desc"
                documents.sort(key=lambda x: x.get(sort) if sort in x else float('inf'), reverse=reverse)

            documents = documents[skip:] if limit is None else documents[skip:skip + limit]

            return documents


    def update(self, collection: str, updates: dict, query: dict =None, limit: int =0) -> list:
        """Update all documents in a collection that match the query."""
        
        with self.LOCK:
            self._validate_collection_exists(collection)
            
            updated_count = 0
            updated_documents = []
            
            if self.get_count(collection) <= 0:
                return updated_documents
            
            if query is None:
                query = {}
                
            db = self._read_db()
            collection_data = db[collection]

            for doc in collection_data:
                if updated_count >= limit and limit > 0:
                    break

                match = True
                for k, v in query.items():
                    if k == '$or' and isinstance(v, list):
                        match = any(
                            all(self._evaluate_condition(doc.get(field), cond) for field, cond in subquery.items())
                            for subquery in v
                        )
                    elif k == '$and' and isinstance(v, list):
                        match = all(
                            all(self._evaluate_condition(doc.get(field), cond) for field, cond in subquery.items())
                            for subquery in v
                        )
                    else:
                        match = self._evaluate_condition(doc.get(k), v)

                    if not match:
                        break

                if match:
                    updated_doc = {**doc, **updates}
                    try:
                        self._validate_document(collection, updated_doc)
                    except DocumentValidationError as e:
                        raise e
                    doc.update(updates)
                    updated_count += 1
                    updated_documents.append(doc)

            self._write_db(db)
            return updated_documents


    def delete(self, collection: str, query: dict = None, limit: int = 0) -> list:
        """Delete documents from a collection that match the query. If no query is provided, delete the first N documents (or all if limit=0)."""

        with self.LOCK:
            self._validate_collection_exists(collection)

            if self.get_count(collection) <= 0:
                return []

            deleted_docs = []
            db = self._read_db()
            collection_data = db[collection]

            if query is None:
                if limit == 0:
                    deleted_docs = collection_data[:]
                    db[collection] = []
                else:
                    deleted_docs = collection_data[:limit]  # ✅ Delete first N
                    db[collection] = collection_data[limit:]
            else:
                new_data = []
                matched_count = 0

                for doc in collection_data:
                    if limit > 0 and matched_count >= limit:
                        new_data.append(doc)
                        continue

                    match = True
                    for k, v in query.items():
                        if k == '$or' and isinstance(v, list):
                            match = any(
                                all(self._evaluate_condition(doc.get(field), cond) for field, cond in subquery.items())
                                for subquery in v
                            )
                        elif k == '$and' and isinstance(v, list):
                            match = all(
                                all(self._evaluate_condition(doc.get(field), cond) for field, cond in subquery.items())
                                for subquery in v
                            )
                        else:
                            match = self._evaluate_condition(doc.get(k), v)

                        if not match:
                            break

                    if match:
                        deleted_docs.append(doc)
                        matched_count += 1
                    else:
                        new_data.append(doc)

                db[collection] = new_data

            self._write_db(db)
            self._set_count(collection)

            return deleted_docs


    def backup_db(self, backup_file: str ="backup") -> str:
        """Create a backup of the database."""
        
        with self.LOCK:

            if not os.path.exists(self.DB_FILE):
                raise FileNotFoundError(f"Database file '{self.DB_FILE}' does not exist.")

            timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
            backup_filename = f"{backup_file}_{timestamp}.json"

            backup_dir = os.path.dirname(backup_file)
            if backup_dir and not os.path.exists(backup_dir):
                os.makedirs(backup_dir)

            try:
                with open(self.DB_FILE, "r") as original, open(backup_filename, "w") as backup:
                    backup.write(original.read())
            except FileNotFoundError:
                raise FileNotFoundError(f"Original database file '{self.DB_FILE}' not found.")
            except Exception as e:
                raise RuntimeError(f"An error occurred during backup: {e}")

            return backup_filename
        
    
    def restore_db(self, backup_file_path: str) -> bool:
        """Restore the database from a backup file."""

        if not backup_file_path.endswith(self.EXT):
            backup_file_path += self.EXT

        with self.LOCK:
            if not os.path.exists(backup_file_path):
                raise FileNotFoundError(f"Backup file '{backup_file_path}' does not exist.")

            with open(backup_file_path, "r") as f:
                backup_db = json.load(f)

            db = self._read_db()

            backup_meta = backup_db.get("_meta", {})
            backup_schemas = backup_meta.get("_schema", {})

            for collection_name, docs in backup_db.items():

                if collection_name == "_meta":
                    continue

                backup_schema = backup_schemas.get(collection_name, {})

                if collection_name not in db:
                    self.collection(collection_name, Utility._string_to_type(backup_schema))

                    for doc in docs:
                        self.add(collection_name, doc)
                    
                    print(f"Restored collection '{collection_name}' with {len(docs)} documents.")

                else:
                    existing_schema = self.get_schema(collection_name)

                    # Convert existing schema to string for comparison
                    existing_schema_str = Utility._type_to_string(existing_schema)

                    existing_normalized = Utility._normalize_schema(existing_schema_str)
                    backup_normalized = Utility._normalize_schema(backup_schema)

                    if existing_normalized == backup_normalized:
                        for doc in docs:
                            self.add(collection_name, doc)
                        print(f"Appended {len(docs)} documents to existing collection '{collection_name}'.")
                    else:
                        new_name = Utility.unique_collection_name(collection_name, db)
                        self.collection(new_name, Utility._string_to_type(backup_schema))

                        for doc in docs:
                            self.add(new_name, doc)

                        print(f"Created new collection '{new_name}' due to schema mismatch. Inserted {len(docs)} documents.")

            print("Restore completed successfully.")
            return True
        return False



import time
import json
import random
import string
import hashlib
from datetime import datetime

from .error import SchemaValidationError


class Utility:
    """Utility class for common operations in Piedb."""
    
    @staticmethod
    def generate_id(collection_name: str, length: int =12) -> str:
        """Generate an UniqueId."""
        
        timestamp = int(time.time())
        timestamp_str = format(timestamp, 'x')[:6]
        
        collection_hash = hashlib.md5(collection_name.encode()).hexdigest()[:3]
        
        random_str = ''.join(random.choices(string.ascii_letters + string.digits, k=6))

        return random_str + collection_hash + timestamp_str

    @staticmethod
    def _type_to_string(schema: dict) -> dict:
        """Convert schema types to string representations and validate supported types."""
    
        def validate_and_convert(d):
            type_map = {"str": str, "int": int, "float": float, "bool": bool, "dict": dict, "list": list, "datetime": datetime}
            
            if not isinstance(d, dict):
                raise SchemaValidationError("Schema must be a dictionary.")
            
            converted = {}
            for k, v in d.items():
                if isinstance(v, dict):
                    converted[k] = validate_and_convert(v)
                elif isinstance(v, type):
                    if v not in type_map.values():
                        raise SchemaValidationError(f"Unsupported type: {v}")
                    converted[k] = v.__name__
                else:
                    raise SchemaValidationError(f"Invalid schema definition for '{k}': {v}")
            return converted

        return validate_and_convert(schema)

    @staticmethod
    def _string_to_type(schema: dict) -> dict:
        """Convert string representations back to types."""
        type_map = {"str": str, "int": int, "float": float, "bool": bool, "dict": dict, "list": list, "datetime": datetime}
        return {k: type_map[v] for k, v in schema.items()}
    
    @staticmethod
    def unique_collection_name(base_name: str, db: dict) -> str:
        """Find a unique collection name by appending an incrementing number."""
        i = 1
        while f"{base_name}_{i}" in db:
            i += 1
        return f"{base_name}_{i}"
        
    @staticmethod
    def _normalize_schema(schema: dict) -> dict:
        """
        Normalize a schema dict:
        - sort keys
        - lowercase type names
        - recursively handle nested dicts
        """

        if not isinstance(schema, dict):
            return schema

        normalized = {}
        for key in sorted(schema.keys()):
            value = schema[key]
            if isinstance(value, dict):
                normalized[key] = Utility._normalize_schema(value)
            elif isinstance(value, str):
                normalized[key] = value.lower()
            else:
                normalized[key] = value
        return normalized

    
class CustomJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, datetime): 
            return obj.isoformat()
        return super().default(obj)
    


class CollectionNotFoundError(Exception):
    """Raised when a requested collection does not exist."""
    
    def __init__(self, collection, message="Collection not found in DB"):
        self.collection = collection
        self.message = f'{message} : {collection}'
        super().__init__(self.message)


class SchemaValidationError(Exception):
    """Raised when a document does not conform to the collection's schema."""
    
    def __init__(self, custom_message, message="Schema validation failed for Document"):
        self.custom_message = custom_message
        self.message = f'{message} : {custom_message}'
        super().__init__(self.message)


class DocumentValidationError(Exception):
    """Raised when a document has invalid fields or missing data."""
    
    def __init__(self,custom_message, message="Data validation failed for Document"):
        self.custom_message = custom_message
        self.message = f'{message} : {custom_message}'
        super().__init__(self.message)


class ReservedKeyError(Exception):
    """Raised when the string contains reserved keys - _schemas and _counts."""
    
    def __init__(self, message="are reserved keys."):
        self.keys = "_schemas and _counts"
        self.message = f"{self.keys} {message}"
        super().__init__(self.message)


class UnsupportedOperatorError(Exception):
    """Raised when an unsupported operator is used in a query."""
    
    def __init__(self, operator, message="is not a supported operator."):
        self.operator = operator
        self.message = f"{operator} {message}"
        super().__init__(self.message)



import os
import json
import shlex
import argparse
from datetime import datetime
from beautifultable import BeautifulTable

from .db import Database

def main():
    
    # Initialize the argument parser
    parser = argparse.ArgumentParser(description="PieDB CLI - A simple JSON database CLI", prog="PieDB CLI")
    
    subparsers = parser.add_subparsers(dest='command')
    
    
    # Database commands
    
    parser_db = subparsers.add_parser('database', help='Database Operations')
    db_subparsers = parser_db.add_subparsers(dest='db_command')

    parser_db_init = db_subparsers.add_parser('init', help='Initialize a new database')
    parser_db_init.add_argument('db_name', type=str, help='The name of the database to initialize')

    parser_db_drop = db_subparsers.add_parser('drop', help='Drop the database')
    parser_db_drop.add_argument('db_name', type=str, help='The name of the database to drop')

    parser_db_list = db_subparsers.add_parser('list', help='List all collections in the database')
    
    # Collection commands
    
    parser_collection = subparsers.add_parser('collection', help='Collection Operations')
    collection_subparsers = parser_collection.add_subparsers(dest='collection_command')
    
    parser_collection_create = collection_subparsers.add_parser('create', help='Create a new collection')
    parser_collection_create.add_argument('collection_name', type=str, help='Name of the collection to create')

    parser_collection_drop = collection_subparsers.add_parser('drop', help='Drop a collection')
    parser_collection_drop.add_argument('collection_name', type=str, help='Name of the collection to drop')

    parser_collection_set_schema = collection_subparsers.add_parser('set_schema', help='Set schema for a collection')
    parser_collection_set_schema.add_argument('collection_name', type=str, help='Name of the collection')
    parser_collection_set_schema.add_argument('schema', type=str, help='Schema for the collection in JSON format')
    
    parser_collection_get_schema = collection_subparsers.add_parser('get_schema', help='Get schema of a collection')
    parser_collection_get_schema.add_argument('collection_name', type=str, help='Name of the collection')
    
    parser_collection_get_data = collection_subparsers.add_parser('get_data', help='Get data from a collection')
    parser_collection_get_data.add_argument('collection_name', type=str, help='Name of the collection')
    
    
    #Document commands
    
    parser_document = subparsers.add_parser('document', help='Document Operations')
    document_subparsers = parser_document.add_subparsers(dest='document_command')
    
    parser_document_add = document_subparsers.add_parser('add', help='Add a document to a collection')
    parser_document_add.add_argument('collection_name', type=str, help='Name of the collection')
    parser_document_add.add_argument('document', type=str, help='Document to add in JSON format')
    
    parser_document_add_many = document_subparsers.add_parser('add_many', help='Add multiple documents to a collection')
    parser_document_add_many.add_argument('collection_name', type=str, help='Name of the collection')
    parser_document_add_many.add_argument('documents', type=str, help='Documents to add in JSON format')
    
    parser_document_update = document_subparsers.add_parser('update', help='Update documents in a collection')
    parser_document_update.add_argument('collection_name', type=str, help='Name of the collection')
    parser_document_update.add_argument('update', type=str, help='Update operations in JSON format')
    parser_document_update.add_argument('--query', type=str, help='Query to match documents')
    parser_document_update.add_argument('--limit', type=int, default=0, help='Update multiple documents')
    
    parser_document_delete = document_subparsers.add_parser('delete', help='Delete documents from a collection')
    parser_document_delete.add_argument('collection_name', type=str, help='Name of the collection')
    parser_document_delete.add_argument('--query', type=str, help='Query to match documents')
    parser_document_delete.add_argument('--limit', type=int, default=0, help='Delete multiple documents')
    
    #Find command
    
    parser_find = subparsers.add_parser('find', help='Find documents in a collection')
    parser_find.add_argument('collection_name', type=str, help='Name of the collection')
    parser_find.add_argument('--query', type=str, default=None, help='Query to match documents')
    parser_find.add_argument('--limit', type=int, default=None, help='Limit the number of documents returned')
    parser_find.add_argument('--skip', type=int, default=0, help='Number of documents to skip')
    parser_find.add_argument('--sort', type=str, default=None, help='Field for sorting the results')
    parser_find.add_argument('--order', type=str, default="asc", help='Sort order for the results')
    
    #Backup command

    parser_backup = subparsers.add_parser('backup', help='Backup the database')
    parser_backup.add_argument('backup_file', type=str, help='Path to the backup file')
    
    parser_restore = subparsers.add_parser('restore', help='Restore the database from a backup')
    parser_restore.add_argument('backup_file', type=str, help='Path to the backup file')
    
    # Info command
    parser_info = subparsers.add_parser("info", help="Show PieDB CLI information")

    print('############### PieDB CLI! ###############\nType "exit" to quit cli.')

    DATABASE = None

    while True:
        try:
            line = input('piedb>> ')
            if not line.strip():
                continue
            if line.strip() in ['exit']:
                print('############### EXIT! ###############')
                break

            if line.strip() == 'clear':
                os.system('cls' if os.name == 'nt' else 'clear')
                continue

            args = parser.parse_args(shlex.split(line))

            if args.command == 'database':
                if args.db_command == 'init':
                    db_name = args.db_name
                    db = Database(db_name)
                    DATABASE = db
                    print(f"Database '{db_name}' initialized.")

                elif args.db_command == 'drop':
                    if DATABASE is None:
                        print("No database initialized. Please initialize a database first.")
                        continue
                    db_name = args.db_name
                    if DATABASE.DB_FILE != db_name + DATABASE.EXT:
                        print(f"Confirm database name to drop: {db_name}")
                        continue
                    result = DATABASE.drop_db()
                    print(f"Database '{db_name}' dropped.")

                elif args.db_command == 'list':
                    if DATABASE is None:
                        print("No database initialized. Please initialize a database first.")
                        continue
                    result = DATABASE.list()
                    print("Collections in the database :")
                    print(json.dumps(result, indent=4))

                else:
                    print('Unknown database subcommand.')
                    
            elif args.command == 'collection':
                if DATABASE is None:
                    print("No database initialized. Please initialize a database first.")
                    continue

                if args.collection_command == 'create':
                    collection_name = args.collection_name
                    DATABASE.collection(collection_name)
                    print(f"Collection '{collection_name}' created.")

                elif args.collection_command == 'drop':
                    collection_name = args.collection_name
                    result = DATABASE.drop_collection(collection_name)
                    print(f"Collection '{collection_name}' dropped.")

                elif args.collection_command == 'set_schema':
                    collection_name = args.collection_name
                    safe_globals = {"__builtins__": {}, "str": str, "int": int, "bool": bool, "float": float, "datetime": datetime}
                    schema = eval(args.schema, safe_globals)
                    DATABASE.set_schema(collection_name, schema)
                    print(f"Schema for collection '{collection_name}' set.")

                elif args.collection_command == 'get_schema':
                    collection_name = args.collection_name
                    schema = DATABASE.get_schema(collection_name)
                    print(f"Schema for collection '{collection_name}':")
                    print(json.dumps(schema, indent=4, default=str))

                elif args.collection_command == 'get_data':
                    collection_name = args.collection_name
                    data = DATABASE.get_collection_data(collection_name)
                    print(f"Data for collection '{collection_name}':")
                    print(json.dumps(data, indent=4))
                    
                else:
                    print('Unknown collection subcommand.')
                    
            elif args.command == 'document':
                if DATABASE is None:
                    print("No database initialized. Please initialize a database first.")
                    continue

                if args.document_command == 'add':
                    collection_name = args.collection_name
                    document = json.loads(args.document)
                    ids = DATABASE.add(collection_name, document)
                    print(f"Document added to collection '{collection_name}': {ids}")

                elif args.document_command == 'add_many':
                    collection_name = args.collection_name
                    documents = eval(args.documents)
                    ids = DATABASE.add_many(collection_name, documents)
                    print(f"Documents added to collection '{collection_name}': {ids}")
                    print(f"{len(documents)} documents added to collection '{collection_name}'.")

                elif args.document_command == 'update':
                    collection_name = args.collection_name
                    update = json.loads(args.update)
                    query = json.loads(args.query)
                    limit = args.limit
                    updated_docs = DATABASE.update(collection_name, update, query, limit)
                    print(f"Updated {len(updated_docs)} documents in collection '{collection_name}'.")

                elif args.document_command == 'delete':
                    collection_name = args.collection_name
                    query = json.loads(args.query)
                    limit = args.limit
                    deleted_docs = DATABASE.delete(collection_name, query, limit)
                    print(f"Deleted {len(deleted_docs)} documents from collection '{collection_name}'.")
                
                else:
                    print('Unknown document subcommand.')
                    
            elif args.command == 'find':
                if DATABASE is None:
                    print("No database initialized. Please initialize a database first.")
                    continue
                
                collection_name = args.collection_name
                query = eval(args.query) if args.query else {}
                limit = args.limit
                skip = args.skip
                sort = args.sort
                order = args.order

                results = DATABASE.find(collection_name, query, limit, skip, sort, order)
                
                if not results:
                    print("No matching documents found.")
                    continue

                table = BeautifulTable()
                table.set_style(BeautifulTable.STYLE_BOX_DOUBLED)
                table.columns.header = ["_id"] + [key for key in results[0] if key != "_id"]

                for doc in results:
                    table.rows.append([doc["_id"]] + [doc[key] for key in doc if key != "_id"])

                print(table)
                
            elif args.command == 'backup':
                if DATABASE is None:
                    print("No database initialized. Please initialize a database first.")
                    continue
                
                backup_file = args.backup_file
                DATABASE.backup_db(backup_file)
                print(f"Database backed up to '{backup_file}'.")

            elif args.command == 'restore':
                if DATABASE is None:
                    print("No database initialized. Please initialize a database first.")
                    continue

                backup_file = args.backup_file
                DATABASE.restore_db(backup_file)
                print(f"Database restored from '{backup_file}'.")
                
            elif args.command == 'info':
                print("PieDB CLI - A simple JSON database CLI")
                print("Version - 2.0.0")
                print("Author - Shubham Kumar Gupta")
                print("For more information, visit: https://pypi.org/project/piedb/")
                print("View on Github - https://github.com/Shubham14243/piedb")
                print("Use 'piedb <command> --help' for more information on a specific command.")

            else:
                print('Unknown Command.')

        except SystemExit:
            pass
        except Exception as e:
            print(f"Error: {str(e)}")

if __name__ == '__main__':
    main()
